/*
 * Copyright (C) 1996-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
__RCSID("$LAAS: cb.c,v 1.22 2003/07/25 09:24:40 matthieu Exp $");

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include <GL/glu.h>
#include <tcl.h>
#include <togl.h>

#include "global.h"
#include "cb.h"
#include "temps.h"

unsigned int FontBase = 0;


static float mat_shininess[] = { 100.0 };

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ((x)*M_PI/180.)
#endif

/*----------------------------------------------------------------------*/

/**
 ** create_cb - appelee apres l'initialisation du widget OpenGL.
 **
 ** definit un certain nombre de parametres par defaut pour OpenGL
 **/
void 
create_cb(const struct Togl *togl)
{
    Tcl_HashEntry *entry;
    int newEntry;
    GDHE_OBS *obs;

    /* printf("create %s\n", Togl_Ident(togl)); */

    if (FontBase == 0) {
	FontBase = Togl_LoadBitmapFont(togl, TOGL_BITMAP_8_BY_13);
	if (!FontBase) {
	    printf("Couldn't load font!\n");
	    exit(1);
	}
    }
    /* parametres par defaut des matieres */
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
    /* Parametres par defaut de l'eclairage */
    glDepthFunc(GL_LEQUAL);

    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

    /* Active l'emploi de glColor pour definir les proprietes des objets */
#if 0
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
#endif

    /* recherche dans la H-table */
    entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
    if (entry == NULL) {
	/* Nouvel nouvel observateur */
	obs = (GDHE_OBS *)Tcl_Alloc(sizeof(GDHE_OBS));
    
	/* Initial viewing mode */
	
	obs->viewMatMode = 0;
	obs->projMode = PERSPECTIVE;
	obs->fovy = 60;
	obs->x = 0;
	obs->y = 0;
	obs->z = 0;
	obs->dist = 10;
	obs->azi = 25;
	obs->elev = 25;
	/* ajout dans la h-table des positions */
	entry = Tcl_CreateHashEntry(posTable, Togl_Ident(togl),
				    &newEntry);
	Tcl_SetHashValue(entry, (ClientData)obs);
    } 
}

/*----------------------------------------------------------------------*/

/**
 ** reshape_cb - appele apres un changement de taille de la fenetre
 **              OpenGL
 **
 ** recalcule la projection
 **/
void 
reshape_cb(const struct Togl *togl)
{
    int width, height;
    GLfloat aspect;

    width = Togl_Width( togl );
    height = Togl_Height( togl );
    aspect = (float) width / (float) height;


    /* nouvelles dimention du viewport */
    glViewport( 0, 0, width, height );
    

}

/*----------------------------------------------------------------------*/

/* 
 * Affichage d'une chaine de caracteres 
 */
static void 
print_string(const char *s)
{
    glCallLists(strlen(s), GL_UNSIGNED_BYTE, s);
}

/*----------------------------------------------------------------------*/

/**
 ** display_cb - appelee pour reafficher le widget OpenGL
 **
 ** Positionne la viewing transformation et appelle 
 ** la fonction TCL 'draw_all' pour faire le sale boulot
 **/
void 
display_cb(const struct Togl *togl)
{
    double maxdist;
    Tcl_Interp *interp = Togl_Interp(togl);
    Tcl_HashEntry *entry;
    GDHE_OBS *obs;
    char buf[1024];
    struct timeval debut, fin;
    unsigned long duree;
    int width, height;
    GLfloat aspect;

    /* enregistrement date dbut  */

    gettimeofday(&debut, NULL);

    /* nettoyage du display */
    
    glClearColor(backgroundColor[0], backgroundColor[1], backgroundColor[2],
		 backgroundColor[3]);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    /* recuperation d'un observateur dans la H-table */

    entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
    if (entry == NULL) {
	 fprintf(stderr, "pb pour trouver la position de togl\n");
	 return;
    }
    obs = (GDHE_OBS *)Tcl_GetHashValue(entry);

    width = Togl_Width( togl );
    height = Togl_Height( togl );
    aspect = (float) width / (float) height;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    /* projection du monde */
    switch (obs->projMode) {
    case CUSTOM:
	 glMultMatrixd(obs->projMat);
	 break;
    case PERSPECTIVE:
	 maxdist = hypot(xmax - xmin, ymax - ymin)*3;
	 gluPerspective(obs->fovy, aspect, 0.01, maxdist);

/* 	 CornerX = -aspect*sin(DEG_TO_RAD(obs->fovy/2))*0.98; */
/* 	 CornerY = -sin(DEG_TO_RAD(obs->fovy/2))*0.98; */
/* 	 CornerZ = -1.0; */
	 break;
    case ORTHOGONAL:
	 maxdist = hypot(xmax - xmin, ymax - ymin)*2;
	 glOrtho(-obs->dist*aspect, obs->dist*aspect, 
		 -obs->dist, obs->dist,  0.01, maxdist);

/* 	 CornerX = -obs->dist*aspect*0.9; */
/* 	 CornerY = -obs->dist*0.9; */
/* 	 CornerZ = -1.0; */
	 break;
    }

    /* point de vue sur le monde */
    if (obs->viewMatMode) {
	 glMultMatrixd(obs->viewMat);
    } else {
	 glTranslated(0.0, 0.0, -obs->dist);    
	 glRotated(obs->elev, 1.0, 0.0, 0.0);
	 glRotated(obs->azi, 0.0, 1.0, 0.0);	 
	 /* Changement de repere GL -> LAAS */
	 glRotated(-90, 1, 0, 0);
	 /* translation */
	 glTranslated(-obs->x, -obs->y, -obs->z);
    }
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    /* les lampes */
    if (enable_lighting) {
      	glEnable(GL_LIGHTING);
    } else {
	glDisable(GL_LIGHTING);
    }
    glEnable( GL_DEPTH_TEST ); 

    /* Appel de la fonction TCL draw_all */
    snprintf(buf, sizeof(buf), "draw_all %s", Tk_PathName(Togl_TkWin(togl))); 
    if (Tcl_Eval(interp, buf) != TCL_OK) {
	Tcl_BackgroundError(interp);
    }

    /* Affichage texte */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-0.5,width-0.5,-0.5,height-0.5,-1.0,1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable( GL_DEPTH_TEST );
    glDisable( GL_LIGHTING );
    glLoadIdentity();
    glColor3f( 1.0, 1.0, 1.0 );
    glRasterPos2i(10,10); 
    glListBase( FontBase );
    
    print_string( "\xa9 1997-2003 LAAS/CNRS" );
    
    Togl_SwapBuffers( togl );

    /*
     * statistiques affichage
     */
    gettimeofday(&fin, NULL);
    duree = time_difference(&fin, &debut);
    snprintf(buf, sizeof(buf), "%ld", duree);
    Tcl_SetVar(interp, "displayTime", buf, TCL_GLOBAL_ONLY);

    /* defini la date du prochain affichage */
    schedule_redisplay(&fin, duree);
    

}

/*----------------------------------------------------------------------*/

/**
 ** Handler des erreurs X
 **/
#ifdef notyet
int
XErrorProc(ClientDate data, XErrorEvent *errEventPtr)
{
    Tk_Window w = (Tk_Window)data;
    
    fprintf(stderr, "GDHE: X protocol Error: ");
    fprintf(stderr, "error=%d request=%d minor=%d\n",
	    errEventPtr->error_code, errEventPtr->request_code,
	    errEventPtr->minor_code);
    return 0;
}
#endif
