# 
# Copyright (C) 1996-2001 LAAS/CNRS 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $LAAS: karma.tcl,v 1.2 2003/05/09 13:48:59 mallet Exp $
#----------------------------------------------------------------------
package provide Gdhe::models 1.0

# Karma (made by Simon in 12/2002)

#----------------------------------------------------------------------
#
# karma
#
# - vectorization : the vectorization angle of the propellers (degrees)
# - frontFOV : display the front cameras FOV (0, 1 or 2 - see camera.tcl)
# - frontFrame : display the front camera frame
# - depthFOV : the depth of the lines (or planes) that defines the FOV
#
#----------------------------------------------------------------------

proc karma {vectorization frontFrame frontFOV depthFOV } {

    # The gondola (a rectangle blue box)
    color 100 100 255
    box 0 0 -0.125 1.0 0.25 0.25

    pushMatrix
    color 0 0 255
    rotate 180 1 0 0 
    repere 1.0
    popMatrix

    # The vectorized propellers
    karmaPropellers $vectorization

    # The stereo bench
    karmaStereoBench $frontFOV $frontFrame $depthFOV

    # The enveloppe
    pushMatrix
    translate 3.0 0.0 1.1
    rotate 180 0 0 1
    enveloppe

    popMatrix
}

# The stereo bench
proc karmaStereoBench {frontFOV frontFrame depthFOV} {

    set hFOV 33.5
    set vFOV 26.5

    color 0 0 0 
    cylinder -1.2 0 0 x 0.03 2.4;	# The axis

    pushMatrix
    translate 1.2 0 0
    rotate 90 0 1 0
    rotate 90 1 0 0
    cameraWithFOV $hFOV $vFOV $frontFOV $depthFOV $frontFrame
    popMatrix
}


# The vectorized propellers
proc karmaPropellers {vectorization} {
    pushMatrix

    translate -0.25 0 0;		# 1/4 of the gondola backward
    rotate $vectorization 0 1 0
    color 0 0 0 
    cylinder 0 -0.5 0 y 0.03 1.0;	# The axis

    pushMatrix
    translate 0 0.5 0 
    color 150 150 150
    cylinder 0 0 0 x 0.3 0.03;		# Left propeller
    popMatrix

    pushMatrix
    translate 0 -0.5 0 
    color 150 150 150
    cylinder 0 0 0 x 0.3 0.03;		# Right propeller
    popMatrix

    popMatrix
}

#----------------------------------------------------------------------
#
# karma internal control parameters
#
#----------------------------------------------------------------------

proc karmaControls {} {
    global frontFrame
    global frontFOV
    global depthFOV

    toplevel .karma
    wm title .karma "Karma controls"
    set dpyWindow .gdhe.ogl

#     wm geometry .karma -100-100

    # The internal configuration parameters
    scale  .karma.vectorization -length 200  -label {Vectorization} -from 80 -to 180 \
	-resolution 1.0 -command {setKarma vectorization } -orient horizontal
    pack .karma.vectorization -fill x

    # The front camera
    # TODO : mettre une barre de sparation
    frame .karma.frontCamera 
    pack .karma.frontCamera -fill x

    label .karma.frontCamera.label -text "Front camera"
    pack .karma.frontCamera.label -fill x

    # Toggle the frame display
    checkbutton .karma.frontCamera.frame -text "Display frame" \
	-variable frontFrame \
    	-command "$dpyWindow redraw" 
    pack .karma.frontCamera.frame -fill x

    # Toggle the field of view display
    set frontFOVmenu \
	[tk_optionMenu .karma.frontCamera.fov frontFOV "No field of view" \
	     "Line FOV" "Filled FOV"]
    .karma.frontCamera.fov configure -width 20
    $frontFOVmenu entryconfigure 0 \
	-command "set frontFOV \"No field of view\"; $dpyWindow redraw"
    $frontFOVmenu entryconfigure 1 \
	-command "set frontFOV \"Line FOV\"; $dpyWindow redraw"
    $frontFOVmenu entryconfigure 2 \
	-command "set frontFOV \"Filled FOV\"; $dpyWindow redraw"
    pack .karma.frontCamera.fov -fill x

    # Changes the the field of view depth
    set depthFOV 10.0;			# TODO : why is this necessary ?
    scale  .karma.frontCamera.depth -length 200  -label {FOV depth (m)} \
	-from 0.0 -to 100.0 -resolution 1.0 \
	-command {setKarma frontCamera.depth } -orient horizontal
    .karma.frontCamera.depth set 10
    pack .karma.frontCamera.depth -fill x

}

# Front camera field of view
# proc setFrontFOV {window} {
#     global pipoFrontFOV
#     global frontFOV

#     set frontFOV $pipoFrontFOV
# }

# Slider events handler
proc setKarma { axis value } {
    global robots
    global depthFOV

    set depthFOV [.karma.frontCamera.depth get]

    set robots(karma) { karma [.karma.vectorization get] }


    .gdhe.ogl redraw
}

### 
### A camera with its field of view
###
proc cameraWithFOV {hFOV vFOV displayFOV depthFOV frame} {
    # hFOV and vFOV : horizontal and vertical half field of view in degrees 
    # displayFOV :	0 -> no display
    #			1 -> display lines
    #			2 -> display planes
    # depthFOV : the depth of the lines (or planes) that defines the FOV
    # frame : draw the camera frame if non 0

    # Camera body definition
    object camera {
	color 120 120 120
	box -0.03 0.0 -0.02 0.06 0.04 0.04
	color 70 70 70
	cylinder 0.0 0.0 0.0 x 0.03 0.01
	cylinder 0.01 0.0 0.0 x 0.03 0.05 0.02
	cylinder 0.03 0.0 0.0 x 0.05 0.01
    }

    # Frame
    if {$frame} {
	pushMatrix
	translate 0 0 0
	color 255 0 0 
	repere 1.0
	popMatrix
    }

    # Field of view
    if {$displayFOV != "No field of view"} {
	color 0 100 0

	set hFOVrad [expr $hFOV * 3.14159 / 180.0]
	set vFOVrad [expr $vFOV * 3.14159 / 180.0]
	set x $depthFOV
	set y [expr $depthFOV * tan($hFOVrad)]
	set z [expr $depthFOV * tan($vFOVrad)]

	# First display the lines
	# "top left" line
	polyline 2 0 0 0 $x $y $z
	# "top right" line
	polyline 2 0 0 0 $x [expr - $y] $z
	# "down right" line
	polyline 2 0 0 0 $x [expr - $y] [expr -$z]
	# "down left" line
	polyline 2 0 0 0 $x $y [expr -$z]
	
	# Down fronto-parallel line
	polyline 2 $x [expr - $y] [expr -$z] $x $y [expr -$z]
	# Top fronto-parallel line
	polyline 2 $x $y $z $x [expr - $y] $z
	# Left fronto-parallel line
	polyline 2 $x $y $z $x $y [expr -$z]
	# Right fronto-parallel line
	polyline 2 $x [expr - $y] $z $x [expr - $y] [expr -$z]

	if {$displayFOV == "Filled FOV"} {		# Planes are also displayed
	    cullFace false
	    color 0 255 0
	    # top plane
	    polygon 3 0 0 0 $x [expr - $y] $z  $x $y $z
	    # bottom plane
	    polygon 3 0 0 0 $x [expr - $y] [expr -$z] $x $y [expr -$z]
	    # left plane
	    polygon 3 0 0 0 $x $y $z $x $y [expr -$z]
	    # right plane
	    polygon 3 0 0 0 $x [expr - $y] $z $x [expr - $y] [expr -$z]
	}
    }
    
}

proc enveloppe {} {
    concave false
    cullFace True

#Corp du dirigeable
    color 150 150 150
    cylinder 0 0 0 x 0 1 0.25
    cylinder 0.25 0 0 x 1 1.5 0.5
    cylinder 0.75 0 0 x 1.5 2 1.25
    
    cylinder 2 0 0 x 2 2 1.5
 
    cylinder 3.5 0 0 x 2 1.25 2.5
    cylinder 6 0 0 x 1.25 0.5 1.75
    cylinder 7.75 0 0 x 0.5 0 0.25

#Ailerons (vue de derriere)
    #haut droit
    pushMatrix
    translate 6.9 0.5 0.5
    rotate 45 1 0 0 
    color 70 70 255
    box 0 0 0 0.8 0.7 0.1 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.1 
    popMatrix


    #haut gauche
    pushMatrix
    translate 6.9 -0.5 0.5
    rotate 135 1 0 0 
    color 70 70 255
    box 0 0 0 0.8 0.7 0.1 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.1 
    popMatrix

    #bas droit
    pushMatrix
    translate 6.9 0.5 -0.5
    rotate -45 1 0 0 
    color 70 70 255
    box 0 0 0 0.8 0.7 0.1 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.1 
    popMatrix
    #bas gauche
    pushMatrix
    translate 6.9 -0.5 -0.5
    rotate -135 1 0 0 
    color 70 70 255
    box 0 0 0 0.8 0.7 0.1 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.1 
    popMatrix

#Surface mobiles

    #haut droit
    pushMatrix
    translate 7.4 0.5 0.5
    rotate 45 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.1 
    popMatrix    

    #haut gauche
    pushMatrix
    translate 7.4 -0.5 0.5
    rotate 135 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.1 
    popMatrix    

    #bas droit
    pushMatrix
    translate 7.4 0.5 -0.5
    rotate -45 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.1 
    popMatrix    

    #bas gauche
    pushMatrix
    translate 7.4 -0.5 -0.5
    rotate -135 1 0 0
    color 70 70 255
    box 0 0 0 0.18 0.7 0.1 
    popMatrix    
}
